from cascade_filter.clause import Clause
from cascade_filter.filter import TextFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable

from typing import cast


class Missing:
	pass


class TextChecker(SingleChecker[TextFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		attr_value = getattr(obj, self.cascade_filter.table_field, Missing)

		if attr_value is Missing:
			return False

		if attr_value is None:
			return self.cascade_filter.clause is Clause.NOT_EQUAL

		attr_value = cast(str, attr_value)

		if self.cascade_filter.clause is Clause.EQUAL:
			return attr_value == self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.NOT_EQUAL:
			return attr_value != self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.CONTAINS:
			return self.cascade_filter.value.value in attr_value
		elif self.cascade_filter.clause is Clause.NOT_CONTAINS:
			return self.cascade_filter.value.value not in attr_value
		elif self.cascade_filter.clause is Clause.STARTS_WITH:
			return attr_value.startswith(self.cascade_filter.value.value)
		elif self.cascade_filter.clause is Clause.ENDS_WITH:
			return attr_value.endswith(self.cascade_filter.value.value)

		raise NotImplementedError(f"Unknown text filter clause: {self.cascade_filter.clause}")



from cascade_filter.field_type import FieldType
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.value import FilterValue

from typing import List, Any


class ChoiceFilterValue(FilterValue[List[Any]]):
	pass


class ChoiceFilter(SingleFilter[List[Any]]):
	subtype = FieldType.CHOICE



from cascade_filter.generic.enum import ExtendedEnum


class FieldType(ExtendedEnum):
	DATE = "date"
	CHOICE = "choice"
	TEXT = "text"
	UUID = "uuid"
	NUMERIC = "numeric"
	BOOLEAN = "boolean"
	ARRAY = "array"



from cascade_filter.field_type import FieldType

from typing import ClassVar


class Field:
	type: ClassVar[FieldType]
	label: str
	nullable: bool

	def __init__(self, label: str, nullable: bool = False) -> None:
		self.label = label
		self.nullable = nullable

	def serialize(self) -> dict:
		return {
			"type": self.type.value,
			"label": self.label,
			"nullable": self.nullable,
		}



from typing import Any, Generic, TypeVar


T = TypeVar("T", bound=Any)


class FilterValue(Generic[T]):
	__slots__ = "label", "value"

	label: str
	value: T

	def __init__(self, label: str, value: T) -> None:
		self.label = label
		self.value = value

	def serialize(self) -> dict:
		return {
			"label": self.label,
			"value": self.value,
		}



from cascade_filter.clause import Clause, CLAUSES_CONTAINING, CLAUSES_NULLABLE
from cascade_filter.exception import ValidationError
from cascade_filter.filter.single.validator import SingleFilterValidator, CleanedData as _CleanedData
from cascade_filter.filter.value.validator import CleanedData as _FilterValueCleanedData

from typing import Final, FrozenSet, List, Any


class FilterValueCleanedData(_FilterValueCleanedData):
	value: List[Any]


class CleanedData(_CleanedData):
	value: FilterValueCleanedData


class ArrayFilterValidator(SingleFilterValidator):
	AVAILABLE_CLAUSES: Final[FrozenSet[Clause]] = frozenset((*CLAUSES_CONTAINING, *CLAUSES_NULLABLE))

	def value(self, value: dict) -> CleanedData:
		base_cleaned_data = super().validate(value)

		if base_cleaned_data["clause"] not in self.AVAILABLE_CLAUSES:
			raise ValidationError("The \"clause\" field must contain one of available clauses")

		if not isinstance(base_cleaned_data["value"]["value"], list):
			raise ValidationError("The \"value\" must be of list type")

		return base_cleaned_data



from cascade_filter.django.widget import CascadeFilterWidget
from cascade_filter.django.validator import CascadeFilterValidator
from cascade_filter.meta import Table

from typing import Type

from django.forms.fields import JSONField


class CascadeFilterField(JSONField):
	"""
	Cascade filter field.
	Allows to represent and edit cascade filter.

	Examples:
		>>> class MyForm(forms.Form):
		>>> 	some_field = CascadeFilterField(table_meta=YourTableMeta)
	"""
	widget: Type[CascadeFilterWidget] = CascadeFilterWidget
	FilterValidator: Type[CascadeFilterValidator] = CascadeFilterValidator

	table_meta: Type[Table]

	def __init__(self, encoder=None, decoder=None, **kwargs):
		table_meta = kwargs.pop(self.widget.ATTR_KEY_TABLE_META)
		assert issubclass(table_meta, Table), f"Table meta must be of {Table.__name__} subclass"
		self.table_meta = table_meta
		kwargs["validators"] = [self.FilterValidator(self.table_meta), *kwargs.get("validators", [])]
		super().__init__(encoder, decoder, **kwargs)

	def widget_attrs(self, widget):
		attrs = super().widget_attrs(widget)
		attrs[self.widget.ATTR_KEY_TABLE_META] = self.table_meta
		return attrs



from cascade_filter.exception import ValidationError
from cascade_filter.clause import Clause
from cascade_filter.field_type import FieldType
from cascade_filter.filter.base import BaseFilterValidator
from cascade_filter.filter.base.validator import CleanedData as _CleanedData
from cascade_filter.filter.single import SingleFilterValidator
from cascade_filter.filter.single.validator import CleanedData as SingleFilterCleanedData
from cascade_filter.filter.type import FilterType
from cascade_filter.filter.date import DateFilterValidator
from cascade_filter.filter.choice import ChoiceFilterValidator
from cascade_filter.filter.text import TextFilterValidator
from cascade_filter.filter.numeric import NumericFilterValidator
from cascade_filter.filter.boolean import BooleanFilterValidator
from cascade_filter.filter.array import ArrayFilterValidator

from typing import List, Union, Final, FrozenSet


class CleanedData(_CleanedData):
	subfilters: List[Union["CleanedData", SingleFilterCleanedData]]


class MultiFilterValidator(BaseFilterValidator):
	AVAILABLE_CLAUSES: Final[FrozenSet[Clause]] = frozenset((Clause.AND, Clause.OR))

	def validate(self, value: dict) -> CleanedData:
		base_cleaned_data = super().validate(value)

		if base_cleaned_data["clause"] not in self.AVAILABLE_CLAUSES:
			raise ValidationError("The \"clause\" field must contain one of available clauses")

		filter_subfilters = value.get("subfilters")

		if not isinstance(filter_subfilters, list):
			raise ValidationError("The \"subfilters\" field must contain valid list")

		single_filter_deserializer = SingleFilterValidator()
		subfilters = []

		for subfilter in filter_subfilters:
			if not isinstance(subfilter, dict):
				raise ValidationError("Each subfilter must be of dict type")

			subfilter_type = subfilter.get("type")

			if not FilterType.has_value(subfilter_type):
				raise ValidationError("Each subfilter must have a valid FilterType in the \"type\" field")

			if subfilter_type == FilterType.SINGLE.value:
				single_filter_cleaned_data = single_filter_deserializer.validate(subfilter)

				if single_filter_cleaned_data["subtype"] is FieldType.DATE:
					validator = DateFilterValidator()
				elif single_filter_cleaned_data["subtype"] is FieldType.CHOICE:
					validator = ChoiceFilterValidator()
				elif single_filter_cleaned_data["subtype"] is FieldType.TEXT:
					validator = TextFilterValidator()
				elif single_filter_cleaned_data["subtype"] is FieldType.NUMERIC:
					validator = NumericFilterValidator()
				elif single_filter_cleaned_data["subtype"] is FieldType.BOOLEAN:
					validator = BooleanFilterValidator()
				elif single_filter_cleaned_data["subtype"] is FieldType.ARRAY:
					validator = ArrayFilterValidator()
				else:
					raise NotImplementedError(
						f"Validation of filters of subtype \"{single_filter_cleaned_data['subtype']}\" is not implemented",
					)

				subfilters.append(validator.validate(subfilter))

			elif subfilter_type == FilterType.MULTI.value:
				subfilters.append(self.validate(subfilter))

			else:
				raise NotImplementedError(
					f"Validation of subfilters of type \"{subfilter_type}\" is not implemented",
				)

		return CleanedData(**base_cleaned_data, subfilters=subfilters)



from cascade_filter.filter import BaseFilter, SingleFilter, MultiFilter
from cascade_filter.django.filter.multi_filter_mixin import MultiFilterMixin

from typing import TypeVar

from django.db.models.base import Model
from django.db.models.query import QuerySet


T = TypeVar("T", bound=Model)


class Filter(MultiFilterMixin):
	cascade_filter: BaseFilter

	def __init__(self, cascade_filter: BaseFilter) -> None:
		self.cascade_filter = cascade_filter

	def filter(self, qs: QuerySet[T]) -> QuerySet[T]:
		if isinstance(self.cascade_filter, SingleFilter):
			q = self.make_single_filter(self.cascade_filter)
		elif isinstance(self.cascade_filter, MultiFilter):
			q = self.make_multi_filter(self.cascade_filter)
		else:
			raise NotImplementedError(f"Unknown filter type: {type(self.cascade_filter)}")

		return qs.filter(q)



from cascade_filter.filter import BaseFilter
from cascade_filter.filter.value import FilterValue
from cascade_filter.clause import Clause
from cascade_filter.meta.field import FieldType
from cascade_filter.filter.type import FilterType

from abc import ABCMeta
from typing import ClassVar, TypeVar, Any, Generic


T = TypeVar("T", bound=Any)


class SingleFilter(BaseFilter, Generic[T], metaclass=ABCMeta):
	__slots__ = BaseFilter.__slots__ + ("table_field", "value")

	type = FilterType.SINGLE
	subtype: ClassVar[FieldType]
	table_field: str
	value: FilterValue[T]

	def __init__(self, clause: Clause, enabled: bool, table_field: str, value: FilterValue[T]):
		super().__init__(clause, enabled)
		self.table_field = table_field
		self.value = value



from cascade_filter.filter import BaseFilter
from cascade_filter.field_type import FieldType
from cascade_filter.filter.type import FilterType
from cascade_filter.filter.multi import MultiFilterDeserializer
from cascade_filter.filter.date import DateFilterDeserializer
from cascade_filter.filter.choice import ChoiceFilterDeserializer
from cascade_filter.filter.text import TextFilterDeserializer
from cascade_filter.filter.numeric import NumericFilterDeserializer
from cascade_filter.filter.boolean import BooleanFilterDeserializer
from cascade_filter.filter.array import ArrayFilterDeserializer
from cascade_filter.filter.uuid import UUIDFilterDeserializer


def deserialize(raw_filter: dict) -> BaseFilter:
	filter_type = raw_filter.get("type")

	if filter_type == FilterType.SINGLE.value:
		filter_subtype = raw_filter.get("subtype")

		if filter_subtype == FieldType.DATE.value:
			deserializer = DateFilterDeserializer()
		elif filter_subtype == FieldType.CHOICE.value:
			deserializer = ChoiceFilterDeserializer()
		elif filter_subtype == FieldType.TEXT.value:
			deserializer = TextFilterDeserializer()
		elif filter_subtype == FieldType.UUID.value:
			deserializer = UUIDFilterDeserializer()
		elif filter_subtype == FieldType.NUMERIC.value:
			deserializer = NumericFilterDeserializer()
		elif filter_subtype == FieldType.BOOLEAN.value:
			deserializer = BooleanFilterDeserializer()
		elif filter_subtype == FieldType.ARRAY.value:
			deserializer = ArrayFilterDeserializer()
		else:
			raise NotImplementedError(f"Deserialization of filters of subtype \"{filter_subtype}\" is not implemented")

	elif filter_type == FilterType.MULTI.value:
		deserializer = MultiFilterDeserializer()

	else:
		raise NotImplementedError(f"Deserialization of filters of type \"{filter_type}\" is not implemented")

	return deserializer.deserialize(raw_filter)		# type: ignore - Validators must do their job



from cascade_filter.filter import UUIDFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class UUIDFilterMixin:
	def make_uuid_filter(self, uuid_filter: UUIDFilter):
		if uuid_filter.clause is Clause.EQUAL:
			return Q(**{uuid_filter.table_field: uuid_filter.value.value})
		elif uuid_filter.clause is Clause.NOT_EQUAL:
			return ~Q(**{uuid_filter.table_field: uuid_filter.value.value})

		raise NotImplementedError(f"Unknown UUID filter clause: \"{uuid_filter.clause}\"")



from cascade_filter.filter.array.dataclass import ArrayFilter
from cascade_filter.filter.array.validator import ArrayFilterValidator
from cascade_filter.filter.array.deserializer import ArrayFilterDeserializer


__all__ = ["ArrayFilter", "ArrayFilterValidator", "ArrayFilterDeserializer"]



from cascade_filter.field_type import FieldType
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.value import FilterValue


class TextFilterValue(FilterValue[str]):
	pass


class TextFilter(SingleFilter[str]):
	subtype = FieldType.TEXT



from cascade_filter.filter.value.dataclass import FilterValue

from typing import TypedDict, TypeVar, Generic, Any


T = TypeVar("T")


class ExpectedValue(TypedDict):
	label: str
	value: Any


class FilterValueDeserializer(Generic[T]):
	def deserialize(self, value: ExpectedValue) -> FilterValue[T]:
		return FilterValue(**value)



from cascade_filter.filter.boolean.dataclass import BooleanFilter
from cascade_filter.filter.boolean.validator import BooleanFilterValidator
from cascade_filter.filter.boolean.deserializer import BooleanFilterDeserializer


__all__ = ["BooleanFilter", "BooleanFilterValidator", "BooleanFilterDeserializer"]



from cascade_filter.filter.base import BaseFilterValidator
from cascade_filter.filter.single import SingleFilterValidator
from cascade_filter.filter.multi import MultiFilterValidator
from cascade_filter.filter.type import FilterType
from cascade_filter.filter.date import DateFilterValidator
from cascade_filter.filter.choice import ChoiceFilterValidator
from cascade_filter.filter.text import TextFilterValidator
from cascade_filter.filter.uuid import UUIDFilterValidator
from cascade_filter.filter.numeric import NumericFilterValidator
from cascade_filter.filter.boolean import BooleanFilterValidator
from cascade_filter.filter.array import ArrayFilterValidator
from cascade_filter.exception import ValidationError
from cascade_filter.meta import Table
from cascade_filter.field_type import FieldType

from json import loads
from logging import getLogger; log = getLogger(__name__)
from typing import Type

from django.core.validators import BaseValidator
from django.core.exceptions import ValidationError


class CascadeFilterValidator(BaseValidator):
	def __init__(self, table_meta: Type[Table]) -> None:
		self.table_meta = table_meta

	def __call__(self, value):
		if isinstance(value, str):
			try:
				value = loads(value)
			except Exception as err:
				log.debug(err)
				raise ValidationError("Invalid JSON-format")

		if not isinstance(value, dict):
			raise ValidationError("Value must be valid JSON-object")

		filter_editor_version = value.get("filterEditorVersion")

		if not isinstance(filter_editor_version, int) or filter_editor_version < 1:
			raise ValidationError("The \"filterEditorVersion\" must be positive integer")

		raw_filter = value.get("filter")

		if not isinstance(raw_filter, dict) and raw_filter is not None:
			raise ValidationError("The \"filter\" field value must be valid JSON-object or null")

		if raw_filter is not None:
			filter_validator = BaseFilterValidator()

			try:
				base_cleaned_data = filter_validator.validate(raw_filter)

				if base_cleaned_data["type"] is FilterType.SINGLE:
					single_filter_validator = SingleFilterValidator()
					single_filter_cleaned_data = single_filter_validator.validate(raw_filter)

					if single_filter_cleaned_data["subtype"] is FieldType.DATE:
						validator = DateFilterValidator()
					elif single_filter_cleaned_data["subtype"] is FieldType.CHOICE:
						validator = ChoiceFilterValidator()
					elif single_filter_cleaned_data["subtype"] is FieldType.TEXT:
						validator = TextFilterValidator()
					elif single_filter_cleaned_data["subtype"] is FieldType.UUID:
						validator = UUIDFilterValidator()
					elif single_filter_cleaned_data["subtype"] is FieldType.NUMERIC:
						validator = NumericFilterValidator()
					elif single_filter_cleaned_data["subtype"] is FieldType.BOOLEAN:
						validator = BooleanFilterValidator()
					elif single_filter_cleaned_data["subtype"] is FieldType.ARRAY:
						validator = ArrayFilterValidator()
					else:
						raise NotImplementedError(
							f"Validation of filters of subtype \"{single_filter_cleaned_data['subtype']}\" is not implemented",
						)

					validator.validate(raw_filter)

				elif base_cleaned_data["type"] is FilterType.MULTI:
					multi_filter_validator = MultiFilterValidator()
					multi_filter_validator.validate(raw_filter)
				else:
					raise NotImplementedError(
						f"Validating filter of type \"{base_cleaned_data['type']}\" is not implemented",
					)

			except ValidationError as err:
				raise ValidationError(*err.args) from err



from cascade_filter.clause import Clause
from cascade_filter.filter.uuid import UUIDFilter
from cascade_filter.filter.value.deserializer import FilterValueDeserializer as _FilterValueDeserializer
from cascade_filter.filter.value.dataclass import FilterValue

from typing import TypedDict
from uuid import UUID


class FilterValueExpectedValue(TypedDict):
	label: str
	value: UUID


class ExpectedValue(TypedDict):
	type: str
	subtype: str
	clause: str
	enabled: bool
	tableField: str
	value: FilterValueExpectedValue


class FilterValueDeserializer(_FilterValueDeserializer[UUID]):
	def deserialize(self, value) -> FilterValue[UUID]:
		return FilterValue(value["label"], UUID(value["value"]))


class UUIDFilterDeserializer:
	def __init__(self) -> None:
		self.value_deserializer = FilterValueDeserializer()

	def deserialize(self, value: ExpectedValue) -> UUIDFilter:
		return UUIDFilter(
			clause=Clause(value["clause"]),
			enabled=value["enabled"],
			table_field=value["tableField"],
			value=self.value_deserializer.deserialize(value["value"]),
		)



from cascade_filter.generic.enum import ExtendedEnum

from typing import Final, FrozenSet


class Clause(ExtendedEnum):
	EQUAL = "=="
	NOT_EQUAL = "!="
	GREATER_THAN = ">"
	LESS_THAN = "<"
	GREATER_THAN_OR_EQUAL = ">="
	LESS_THAN_OR_EQUAL = "<="
	CONTAINS = "contains"
	NOT_CONTAINS = "not contains"
	IS = "is"
	IS_NOT = "is not"
	STARTS_WITH = "starts with"
	ENDS_WITH = "ends with"
	AND = "and"
	OR = "or"
	IS_NULL = "is null"
	IS_NOT_NULL = "is not null"


CLAUSES_STRICT_COMPARISON: Final[FrozenSet[Clause]] = frozenset((Clause.EQUAL, Clause.NOT_EQUAL))
CLAUSES_NON_STRICT_COMPARISON: Final[FrozenSet[Clause]] = frozenset((
	Clause.GREATER_THAN, Clause.LESS_THAN, Clause.GREATER_THAN_OR_EQUAL, Clause.LESS_THAN_OR_EQUAL,
))
CLAUSES_CONTAINING: Final[FrozenSet[Clause]] = frozenset((Clause.CONTAINS, Clause.NOT_CONTAINS))
CLAUSES_ACCORDANCE: Final[FrozenSet[Clause]] = frozenset((Clause.IS, Clause.IS_NOT))
CLAUSES_BOUNDARY: Final[FrozenSet[Clause]] = frozenset((Clause.STARTS_WITH, Clause.ENDS_WITH))
CLAUSES_NULLABLE: Final[FrozenSet[Clause]] = frozenset((Clause.IS_NULL, Clause.IS_NOT_NULL))



from cascade_filter.clause import Clause
from cascade_filter.filter import MultiFilter
from cascade_filter.compliance_checker.base import BaseChecker

from typing import List


class MultiChecker(BaseChecker):
	cascade_filter: MultiFilter

	def __init__(self, cascade_filter: MultiFilter, subcheckers: List[BaseChecker]) -> None:
		super().__init__(cascade_filter)
		self.subcheckers = subcheckers

	def is_fit(self, obj: object) -> bool:
		if self.cascade_filter.clause is Clause.AND:
			return all((i.is_fit(obj) for i in self.subcheckers))
		elif self.cascade_filter.clause is Clause.OR:
			return any((i.is_fit(obj) for i in self.subcheckers))

		raise NotImplementedError(f"Unknown multi filter clause: {self.cascade_filter.clause}")



from cascade_filter.field_type import FieldType
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.value import FilterValue

from uuid import UUID


class UUIDFilterValue(FilterValue[UUID]):
	pass


class UUIDFilter(SingleFilter[UUID]):
	subtype = FieldType.UUID



from cascade_filter.filter import TextFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class TextFilterMixin:
	def make_text_filter(self, text_filter: TextFilter):
		if text_filter.clause is Clause.EQUAL:
			return Q(**{text_filter.table_field: text_filter.value.value})
		elif text_filter.clause is Clause.NOT_EQUAL:
			return ~Q(**{text_filter.table_field: text_filter.value.value})
		elif text_filter.clause is Clause.CONTAINS:
			return Q(**{f"{text_filter.table_field}__contains": text_filter.value.value})
		elif text_filter.clause is Clause.NOT_CONTAINS:
			return ~Q(**{f"{text_filter.table_field}__contains": text_filter.value.value})
		elif text_filter.clause is Clause.STARTS_WITH:
			return Q(**{f"{text_filter.table_field}__startswith": text_filter.value.value})
		elif text_filter.clause is Clause.ENDS_WITH:
			return Q(**{f"{text_filter.table_field}__endswith": text_filter.value.value})

		raise NotImplementedError(f"Unknown text filter clause: \"{text_filter.clause}\"")



from cascade_filter.filter.type import FilterType
from cascade_filter.clause import Clause

from abc import ABC
from typing import ClassVar


class BaseFilter(ABC):
	__slots__ = "clause", "enabled"

	type: ClassVar[FilterType]
	clause: Clause
	enabled: bool

	def __init__(self, clause: Clause, enabled: bool) -> None:
		self.clause = clause
		self.enabled = enabled



from cascade_filter.meta import Table

from typing import Type, Final
from json import loads

from django.forms.widgets import Widget


class CascadeFilterWidget(Widget):
	"""
	Cascade filter editor widget.

	Notes:
		Use only in conjunction with the JSON field and only if reliability is not needed.

	Examples:
		>>> class SomeForm(forms.Form):
		>>> 	some_field = forms.JSONField(
		>>> 		widget=CascadeFilterWidget(
		>>> 			attrs={'table_meta': YourTableMeta}
		>>> 		)
		>>> 	)
	"""
	ATTR_KEY_TABLE_META: Final[str] = "table_meta"
	template_name = "cascade_filter/widget.html"

	class Media:
		css = {
			"all": ("cascade_filter/cascade-filter.css", ),
		}

	def format_value(self, value):
		if isinstance(value, str) and value.startswith("\""):
			return str(loads(value))		# allows to use with django-annoying JSON field

		return value

	def extract_table_meta(self) -> Type[Table]:
		table_meta = self.attrs.get(self.ATTR_KEY_TABLE_META)

		if not isinstance(table_meta, type):
			raise TypeError("The table meta must be class")

		if not issubclass(table_meta, Table):
			raise TypeError(f"Table meta must be subclass of {Table.__name__}")

		return table_meta

	def build_attrs(self, base_attrs, extra_attrs=None):
		attrs = super().build_attrs(base_attrs, extra_attrs)

		if self.ATTR_KEY_TABLE_META in attrs:		# Removing table_meta because it's not serializable
			del attrs[self.ATTR_KEY_TABLE_META]

		return attrs

	def get_context(self, name, value, attrs):
		context = super().get_context(name, value, attrs)
		context[self.ATTR_KEY_TABLE_META] = self.extract_table_meta().serialize()
		return context



from cascade_filter.clause import Clause
from cascade_filter.filter.base import BaseFilter
from cascade_filter.filter.type import FilterType

from typing import List


class MultiFilter(BaseFilter):
	__slots__ = BaseFilter.__slots__ + ("subfilters", )

	type = FilterType.MULTI
	subfilters: List[BaseFilter]

	def __init__(self, clause: Clause, enabled: bool, subfilters: List[BaseFilter]) -> None:
		super().__init__(clause, enabled)
		self.subfilters = subfilters



from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.clause import Clause

from typing import Callable, TypeVar
from functools import wraps


T = TypeVar("T", bound=SingleChecker)
Method = Callable[[T, object], bool]


class Missing:
	pass


def nullable(method: Method):
	@wraps(method)
	def inner(checker: SingleChecker, obj: object) -> bool:
		attr_value = getattr(obj, checker.cascade_filter.table_field, Missing)

		if attr_value is Missing:
			return False

		if checker.cascade_filter.clause is Clause.IS_NULL:
			return attr_value is None
		elif checker.cascade_filter.clause is Clause.IS_NOT_NULL:
			return attr_value is not None

		return method(checker, obj)

	return inner



from cascade_filter.clause import Clause
from cascade_filter.filter import NumericFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable

from typing import cast


class Missing:
	pass


class NumericChecker(SingleChecker[NumericFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		attr_value = getattr(obj, self.cascade_filter.table_field, Missing)

		if attr_value is Missing:
			return False

		if attr_value is None:
			return self.cascade_filter.clause is Clause.NOT_EQUAL

		attr_value = cast(int, attr_value)

		if self.cascade_filter.clause is Clause.EQUAL:
			return attr_value == self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.NOT_EQUAL:
			return attr_value != self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.GREATER_THAN:
			return attr_value > self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.LESS_THAN:
			return attr_value < self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.GREATER_THAN_OR_EQUAL:
			return attr_value >= self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.LESS_THAN_OR_EQUAL:
			return attr_value <= self.cascade_filter.value.value

		raise NotImplementedError(f"Unknown number filter clause: {self.cascade_filter.clause}")



from cascade_filter.filter.uuid.dataclass import UUIDFilter
from cascade_filter.filter.uuid.validator import UUIDFilterValidator
from cascade_filter.filter.uuid.deserializer import UUIDFilterDeserializer


__all__ = ["UUIDFilter", "UUIDFilterValidator", "UUIDFilterDeserializer"]



from cascade_filter.field_type import FieldType
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.value import FilterValue


class NumericFilterValue(FilterValue[int]):
	pass


class NumericFilter(SingleFilter[int]):
	subtype = FieldType.NUMERIC



from cascade_filter.filter.value.dataclass import FilterValue
from cascade_filter.filter.value.validator import FilterValueValidator
from cascade_filter.filter.value.deserializer import FilterValueDeserializer


__all__ = ["FilterValue", "FilterValueValidator", "FilterValueDeserializer"]



from cascade_filter.filter import MultiFilter, SingleFilter
from cascade_filter.clause import Clause
from cascade_filter.django.filter.single_filter_mixin import SingleFilterMixin

from django.db.models.query import Q


class MultiFilterMixin(SingleFilterMixin):
	def make_multi_filter(self, multi_filter: MultiFilter) -> Q:
		if not multi_filter.enabled:
			return Q()

		q = Q()

		for subfilter in multi_filter.subfilters:
			if isinstance(subfilter, SingleFilter):
				sub_q = self.make_single_filter(subfilter)
			elif isinstance(subfilter, MultiFilter):
				sub_q = self.make_multi_filter(subfilter)
			else:
				raise NotImplementedError(f"Unknown filter type: {type(subfilter)}")

			if multi_filter.clause is Clause.AND:
				q &= sub_q
			elif multi_filter.clause is Clause.OR:
				q |= sub_q
			else:
				raise NotImplementedError(f"Unknown multi-filter clause: \"{multi_filter.clause}\"")

		return q



from cascade_filter.meta.field.field import Field
from cascade_filter.field_type import FieldType
from cascade_filter.meta.field.array_field import ArrayField
from cascade_filter.meta.field.boolean_field import BooleanField
from cascade_filter.meta.field.choice_field import ChoiceField
from cascade_filter.meta.field.date_field import DateField
from cascade_filter.meta.field.numeric_field import NumericField
from cascade_filter.meta.field.text_field import TextField
from cascade_filter.meta.field.uuid_field import UUIDField


__all__ = [
	"Field", "FieldType", "ArrayField", "BooleanField", "ChoiceField", "DateField", "NumericField", "TextField",
	"UUIDField",
]



from cascade_filter.filter.date.dataclass import DateFilter
from cascade_filter.filter.date.validator import DateFilterValidator
from cascade_filter.filter.date.deserializer import DateFilterDeserializer


__all__ = ["DateFilter", "DateFilterValidator", "DateFilterDeserializer"]



from cascade_filter.filter import DateFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class DateFilterMixin:
	def make_date_filter(self, date_filter: DateFilter) -> Q:
		if date_filter.clause is Clause.EQUAL:
			return Q(**{date_filter.table_field: date_filter.value.value})
		elif date_filter.clause is Clause.NOT_EQUAL:
			return ~Q(**{f"{date_filter.table_field}": date_filter.value.value})
		elif date_filter.clause is Clause.GREATER_THAN:
			return Q(**{f"{date_filter.table_field}__gt": date_filter.value.value})
		elif date_filter.clause is Clause.LESS_THAN:
			return Q(**{f"{date_filter.table_field}__lt": date_filter.value.value})
		elif date_filter.clause is Clause.GREATER_THAN_OR_EQUAL:
			return Q(**{f"{date_filter.table_field}__gte": date_filter.value.value})
		elif date_filter.clause is Clause.LESS_THAN_OR_EQUAL:
			return Q(**{f"{date_filter.table_field}__lte": date_filter.value.value})

		raise NotImplementedError(f"Unknown date filter clause: \"{date_filter.clause}\"")



from cascade_filter.filter import BaseFilter

from abc import ABC, abstractmethod


class BaseChecker(ABC):
	cascade_filter: BaseFilter

	def __init__(self, cascade_filter: BaseFilter) -> None:
		self.cascade_filter = cascade_filter

	@abstractmethod
	def is_fit(self, obj: object) -> bool: ...



from cascade_filter.filter.base import BaseFilter
from cascade_filter.filter.date import DateFilter
from cascade_filter.filter.choice import ChoiceFilter
from cascade_filter.filter.text import TextFilter
from cascade_filter.filter.uuid import UUIDFilter
from cascade_filter.filter.numeric import NumericFilter
from cascade_filter.filter.boolean import BooleanFilter
from cascade_filter.filter.array import ArrayFilter
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.multi import MultiFilter


__all__ = [
	"BaseFilter", "DateFilter", "ChoiceFilter", "TextFilter", "UUIDFilter", "NumericFilter", "BooleanFilter",
	"ArrayFilter", "SingleFilter", "MultiFilter",
]



from cascade_filter.filter import ArrayFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class ArrayFilterMixin:
	def make_array_filter(self, array_filter: ArrayFilter) -> Q:
		if array_filter.clause is Clause.CONTAINS:
			return Q(**{f"{array_filter.table_field}__contains": array_filter.value.value})
		elif array_filter.clause is Clause.NOT_CONTAINS:
			return ~Q(**{f"{array_filter.table_field}__contains": array_filter.value.value})

		raise NotImplementedError(f"Unknown array filter clause: \"{array_filter.clause}\"")



from cascade_filter.filter.choice.dataclass import ChoiceFilter
from cascade_filter.filter.choice.validator import ChoiceFilterValidator
from cascade_filter.filter.choice.deserializer import ChoiceFilterDeserializer


__all__ = ["ChoiceFilter", "ChoiceFilterValidator", "ChoiceFilterDeserializer"]



from cascade_filter.clause import Clause
from cascade_filter.filter.array.dataclass import ArrayFilter
from cascade_filter.filter.value.deserializer import FilterValueDeserializer

from typing import TypedDict, List, Any


class FilterValueExpectedValue(TypedDict):
	label: str
	value: List[Any]


class ExpectedValue(TypedDict):
	type: str
	subtype: str
	clause: str
	enabled: bool
	tableField: str
	value: FilterValueExpectedValue


class ArrayFilterDeserializer:
	def __init__(self) -> None:
		self.value_deserializer = FilterValueDeserializer[List[Any]]()

	def deserialize(self, value: ExpectedValue) -> ArrayFilter:
		return ArrayFilter(
			clause=Clause(value["clause"]),
			enabled=value["enabled"],
			table_field=value["tableField"],
			value=self.value_deserializer.deserialize(value["value"]),
		)



from cascade_filter.django.filter.text_filter_mixin import TextFilterMixin
from cascade_filter.django.filter.date_filter_mixin import DateFilterMixin
from cascade_filter.django.filter.choice_filter_mixin import ChoiceFilterMixin
from cascade_filter.django.filter.numeric_filter_mixin import NumericFilterMixin
from cascade_filter.django.filter.boolean_filter_mixin import BooleanFilterMixin
from cascade_filter.django.filter.array_filter_mixin import ArrayFilterMixin
from cascade_filter.django.filter.uuid_filter_mixin import UUIDFilterMixin
from cascade_filter.filter import (
	SingleFilter, DateFilter, ChoiceFilter, TextFilter, NumericFilter, BooleanFilter, ArrayFilter, UUIDFilter,
)
from cascade_filter.clause import Clause

from django.db.models.query import Q


class SingleFilterMixin(
	TextFilterMixin, DateFilterMixin, ChoiceFilterMixin, NumericFilterMixin, BooleanFilterMixin, ArrayFilterMixin,
	UUIDFilterMixin,
):
	def make_single_filter(self, single_filter: SingleFilter) -> Q:
		if not single_filter.enabled:
			return Q()

		if single_filter.clause is Clause.IS_NULL:
			return Q(**{f"{single_filter.table_field}__isnull": True})
		elif single_filter.clause is Clause.IS_NOT_NULL:
			return Q(**{f"{single_filter.table_field}__isnull": False})

		if isinstance(single_filter, DateFilter):
			return self.make_date_filter(single_filter)
		elif isinstance(single_filter, ChoiceFilter):
			return self.make_choice_filter(single_filter)
		elif isinstance(single_filter, TextFilter):
			return self.make_text_filter(single_filter)
		elif isinstance(single_filter, NumericFilter):
			return self.make_numeric_filter(single_filter)
		elif isinstance(single_filter, BooleanFilter):
			return self.make_boolean_filter(single_filter)
		elif isinstance(single_filter, ArrayFilter):
			return self.make_array_filter(single_filter)
		elif isinstance(single_filter, UUIDFilter):
			return self.make_uuid_filter(single_filter)

		raise NotImplementedError(f"Unknown filter subtype: \"{single_filter.subtype}\"")



from cascade_filter.filter.text.dataclass import TextFilter
from cascade_filter.filter.text.validator import TextFilterValidator
from cascade_filter.filter.text.deserializer import TextFilterDeserializer


__all__ = ["TextFilter", "TextFilterValidator", "TextFilterDeserializer"]



from cascade_filter.filter import (
	BaseFilter, SingleFilter, MultiFilter, DateFilter, ChoiceFilter, TextFilter, NumericFilter, BooleanFilter,
	ArrayFilter, UUIDFilter,
)
from cascade_filter.compliance_checker.base import BaseChecker
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.multi import MultiChecker
from cascade_filter.compliance_checker.date import DateChecker
from cascade_filter.compliance_checker.choice import ChoiceChecker
from cascade_filter.compliance_checker.text import TextChecker
from cascade_filter.compliance_checker.numeric import NumericChecker
from cascade_filter.compliance_checker.boolean import BooleanChecker
from cascade_filter.compliance_checker.array import ArrayChecker
from cascade_filter.compliance_checker.uuid import UUIDChecker


class ComplianceChecker:
	"""
	Allows to check the compliance of objects with the rules of cascade filters.
	"""
	checker: BaseChecker

	def __init__(self, cascade_filter: BaseFilter) -> None:
		if isinstance(cascade_filter, SingleFilter):
			self.checker = self.make_single_checker(cascade_filter)
		elif isinstance(cascade_filter, MultiFilter):
			self.checker = self.make_multi_checker(cascade_filter)
		else:
			raise NotImplementedError(f"Unknown filter type: {type(cascade_filter)}")

	def is_fit(self, obj: object) -> bool:
		if not self.checker.cascade_filter.enabled:
			return True

		return self.checker.is_fit(obj)

	@classmethod
	def make_single_checker(cls, cascade_filter: SingleFilter) -> SingleChecker:
		if isinstance(cascade_filter, DateFilter):
			return DateChecker(cascade_filter)
		elif isinstance(cascade_filter, ChoiceFilter):
			return ChoiceChecker(cascade_filter)
		elif isinstance(cascade_filter, TextFilter):
			return TextChecker(cascade_filter)
		elif isinstance(cascade_filter, NumericFilter):
			return NumericChecker(cascade_filter)
		elif isinstance(cascade_filter, BooleanFilter):
			return BooleanChecker(cascade_filter)
		elif isinstance(cascade_filter, ArrayFilter):
			return ArrayChecker(cascade_filter)
		elif isinstance(cascade_filter, UUIDFilter):
			return UUIDChecker(cascade_filter)

		raise NotImplementedError(f"Unknown filter type: {type(cascade_filter)}")

	@classmethod
	def make_multi_checker(cls, cascade_filter: MultiFilter) -> MultiChecker:
		subcheckers = []

		for subfilter in cascade_filter.subfilters:
			if not subfilter.enabled:
				continue

			if isinstance(subfilter, SingleFilter):
				subcheckers.append(cls.make_single_checker(subfilter))
			elif isinstance(subfilter, MultiFilter):
				subcheckers.append(cls.make_multi_checker(subfilter))
			else:
				raise NotImplementedError(f"Unknown filter type: {type(cascade_filter)}")

		return MultiChecker(cascade_filter, subcheckers)



from cascade_filter.filter import NumericFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class NumericFilterMixin:
	def make_numeric_filter(self, numeric_filter: NumericFilter) -> Q:
		if numeric_filter.clause is Clause.EQUAL:
			return Q(**{numeric_filter.table_field: numeric_filter.value.value})
		elif numeric_filter.clause is Clause.NOT_EQUAL:
			return ~Q(**{f"{numeric_filter.table_field}": numeric_filter.value.value})
		elif numeric_filter.clause is Clause.GREATER_THAN:
			return Q(**{f"{numeric_filter.table_field}__gt": numeric_filter.value.value})
		elif numeric_filter.clause is Clause.LESS_THAN:
			return Q(**{f"{numeric_filter.table_field}__lt": numeric_filter.value.value})
		elif numeric_filter.clause is Clause.GREATER_THAN_OR_EQUAL:
			return Q(**{f"{numeric_filter.table_field}__gte": numeric_filter.value.value})
		elif numeric_filter.clause is Clause.LESS_THAN_OR_EQUAL:
			return Q(**{f"{numeric_filter.table_field}__lte": numeric_filter.value.value})

		raise NotImplementedError(f"Unknown numeric filter clause: \"{numeric_filter.clause}\"")



from cascade_filter.clause import Clause
from cascade_filter.filter import UUIDFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable

from typing import cast
from uuid import UUID


class Missing:
	pass


class UUIDChecker(SingleChecker[UUIDFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		attr_value = getattr(obj, self.cascade_filter.table_field, Missing)

		if attr_value is Missing:
			return False

		if attr_value is None:
			return self.cascade_filter.clause is Clause.NOT_EQUAL

		attr_value = cast(UUID, attr_value)

		if self.cascade_filter.clause is Clause.EQUAL:
			return attr_value == self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.NOT_EQUAL:
			return attr_value != self.cascade_filter.value.value

		raise NotImplementedError(f"Unknown UUID filter clause: {self.cascade_filter.clause}")



from cascade_filter.meta.field import Field

from typing import List


class Table:
	"""
	Contains info about table fields for filtering table data.

	Examples:
		>>> # your_table.py
		>>> class YourTable(Table):
		>>> 	field1 = TextField('Field 1 Name', nullable=False)
		>>>
		>>> # table_form_meta.py
		>>> class MyTableForm(forms.Form):
		>>> 	field = CascadeFilterField(table_meta=YourTableMeta)
	"""
	@classmethod
	def serialize(cls) -> List[dict]:
		data = []

		for attr_name, attr in cls.__dict__.items():
			if isinstance(attr, Field):
				field_data = attr.serialize()
				field_data["name"] = attr_name
				data.append(field_data)

		return data



from cascade_filter.meta.field.field import Field
from cascade_filter.field_type import FieldType
from cascade_filter.meta.choice import Choice

from typing import Iterable, Union, Callable


ChoicesIterable = Iterable[Choice]
ChoicesCallable = Callable[..., ChoicesIterable]
Choices = Union[ChoicesIterable, ChoicesCallable]


class ChoiceField(Field):
	type = FieldType.CHOICE
	choices: Choices

	def __init__(self, label: str, choices: Choices, nullable: bool = False) -> None:
		super().__init__(label, nullable)
		self.choices = choices

	def serialize(self) -> dict:
		serialized_data = super().serialize()

		if callable(self.choices):
			serialized_data["choices"] = [i.serialize() for i in self.choices()]
		else:
			serialized_data["choices"] = [i.serialize() for i in self.choices]

		return serialized_data



from cascade_filter.clause import Clause
from cascade_filter.filter import BooleanFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable

from typing import cast


class BooleanChecker(SingleChecker[BooleanFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		attr_value = getattr(obj, self.cascade_filter.table_field, None)

		if attr_value is None:
			return False

		attr_value = cast(bool, attr_value)

		if self.cascade_filter.clause is Clause.IS:
			return attr_value == self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.IS_NOT:
			return attr_value != self.cascade_filter.value.value

		raise NotImplementedError(f"Unknown boolean filter clause: {self.cascade_filter.clause}")



from cascade_filter.field_type import FieldType
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.value import FilterValue


class BooleanFilterValue(FilterValue[bool]):
	pass


class BooleanFilter(SingleFilter[bool]):
	subtype = FieldType.BOOLEAN



from cascade_filter.field_type import FieldType
from cascade_filter.exception import ValidationError
from cascade_filter.filter.value import FilterValueValidator
from cascade_filter.filter.value.validator import CleanedData as FilterValueCleanedData
from cascade_filter.filter.base import BaseFilterValidator
from cascade_filter.filter.base.validator import CleanedData as _CleanedData


class CleanedData(_CleanedData):
	subtype: FieldType
	table_field: str
	value: FilterValueCleanedData


class SingleFilterValidator(BaseFilterValidator):
	def validate(self, value: dict) -> CleanedData:
		base_cleaned_data = super().validate(value)
		filter_subtype = value.get("subtype")

		if not FieldType.has_value(filter_subtype):
			raise ValidationError("The \"subtype\" field must contain a valid FieldType")

		filter_table_field = value.get("tableField")

		if not isinstance(filter_table_field, str):
			raise ValidationError("The \"tableField\" field must contain a valid string")

		filter_raw_value = value.get("value")
		filter_value_deserializer = FilterValueValidator()
		filter_value = filter_value_deserializer.validate(filter_raw_value)

		return CleanedData(
			**base_cleaned_data,
			subtype=FieldType(filter_subtype),
			table_field=filter_table_field,
			value=filter_value,
		)



from cascade_filter.clause import Clause
from cascade_filter.filter import ChoiceFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable


class Missing:
	pass


class ChoiceChecker(SingleChecker[ChoiceFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		record_attr_value = getattr(obj, self.cascade_filter.table_field, Missing)

		if record_attr_value is Missing:
			return False

		if self.cascade_filter.clause is Clause.EQUAL:
			return record_attr_value in self.cascade_filter.value.value
		elif self.cascade_filter.clause is Clause.NOT_EQUAL:
			return record_attr_value not in self.cascade_filter.value.value

		raise NotImplementedError(f"Unknown choice filter clause: {self.cascade_filter.clause}")



from cascade_filter.exception import ValidationError

from typing import Any, TypedDict


class CleanedData(TypedDict):
	label: str
	value: Any


class Missing:
	pass


class FilterValueValidator:
	def validate(self, value: Any) -> CleanedData:
		if not isinstance(value, dict):
			raise ValidationError("The filter value must be of dict type")

		filter_label = value.get("label")

		if not isinstance(filter_label, str):
			raise ValidationError("The \"label\" field must be valid string")

		filter_value = value.get("value", Missing)

		if filter_value is Missing:
			raise ValidationError("The \"value\" field is required")

		return CleanedData(label=filter_label, value=filter_value)



from cascade_filter.filter import ChoiceFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class ChoiceFilterMixin:
	def make_choice_filter(self, choice_filter: ChoiceFilter) -> Q:
		if choice_filter.clause is Clause.EQUAL:
			q = Q()

			for option in choice_filter.value.value:
				q |= Q(**{choice_filter.table_field: option})

			return q

		elif choice_filter.clause is Clause.NOT_EQUAL:
			q = Q()

			for option in choice_filter.value.value:
				q &= ~Q(**{choice_filter.table_field: option})

			return q

		raise NotImplementedError(f"Unknown choice filter clause: \"{choice_filter.clause}\"")



from cascade_filter.filter import BooleanFilter
from cascade_filter.clause import Clause

from django.db.models.query import Q


class BooleanFilterMixin:
	def make_boolean_filter(self, boolean_filter: BooleanFilter) -> Q:
		if boolean_filter.clause is Clause.IS:
			return Q(**{boolean_filter.table_field: boolean_filter.value.value})
		elif boolean_filter.clause is Clause.IS_NOT:
			return Q(**{boolean_filter.table_field: not boolean_filter.value.value})

		raise NotImplementedError(f"Unknown boolean filter clause: \"{boolean_filter.clause}\"")



from cascade_filter.clause import Clause
from cascade_filter.exception import ValidationError
from cascade_filter.filter.type import FilterType

from typing import TypedDict


class CleanedData(TypedDict):
	clause: Clause
	enabled: bool
	type: FilterType


class BaseFilterValidator:
	def validate(self, value: dict) -> CleanedData:
		filter_clause = value.get("clause")

		if not Clause.has_value(filter_clause):
			raise ValidationError("The \"clause\" field must contain a valid Clause")

		filter_enabled = value.get("enabled")

		if not isinstance(filter_enabled, bool):
			raise ValidationError("The \"enabled\" field must contain a valid bool")

		filter_type = value.get("type")

		if not FilterType.has_value(filter_type):
			raise ValidationError("The \"enabled\" field must contain a valid FilterType")

		return CleanedData(clause=Clause(filter_clause), enabled=filter_enabled, type=FilterType(filter_type))



from cascade_filter.clause import Clause
from cascade_filter.filter import DateFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable

from arrow import get as get_arrow


class Missing:
	pass


class DateChecker(SingleChecker[DateFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		attr_value = getattr(obj, self.cascade_filter.table_field, Missing)

		if attr_value is Missing:
			return False

		if attr_value is None:
			return self.cascade_filter.clause is Clause.NOT_EQUAL

		attr_value = get_arrow(attr_value)
		filter_value = get_arrow(self.cascade_filter.value.value)

		if self.cascade_filter.clause is Clause.EQUAL:
			return attr_value.date() == filter_value.date()
		elif self.cascade_filter.clause is Clause.NOT_EQUAL:
			return attr_value.date() != filter_value.date()
		elif self.cascade_filter.clause is Clause.GREATER_THAN:
			return attr_value.date() > filter_value.date()
		elif self.cascade_filter.clause is Clause.LESS_THAN:
			return attr_value.date() < filter_value.date()
		elif self.cascade_filter.clause is Clause.GREATER_THAN_OR_EQUAL:
			return attr_value.date() >= filter_value.date()
		elif self.cascade_filter.clause is Clause.LESS_THAN_OR_EQUAL:
			return attr_value.date() <= filter_value.date()

		raise NotImplementedError(f"Unknown date filter clause: {self.cascade_filter.clause}")



from cascade_filter.clause import Clause
from cascade_filter.field_type import FieldType
from cascade_filter.filter.type import FilterType
from cascade_filter.filter.multi.dataclass import MultiFilter
from cascade_filter.filter.value.deserializer import ExpectedValue as FilterValueExpectedValue
from cascade_filter.filter.date import DateFilterDeserializer
from cascade_filter.filter.choice import ChoiceFilterDeserializer
from cascade_filter.filter.text import TextFilterDeserializer
from cascade_filter.filter.numeric import NumericFilterDeserializer
from cascade_filter.filter.boolean import BooleanFilterDeserializer
from cascade_filter.filter.array import ArrayFilterDeserializer

from typing import TypedDict, List, Union, cast


class ExpectedSingleFilterValue(TypedDict):
	type: str
	subtype: str
	clause: str
	enabled: bool
	table_field: str
	value: FilterValueExpectedValue


class ExpectedValue(TypedDict):
	type: str
	clause: str
	enabled: bool
	subfilters: List[Union["ExpectedValue", ExpectedSingleFilterValue]]


class MultiFilterDeserializer:
	def __init__(self) -> None:
		pass

	def deserialize(self, value: ExpectedValue) -> MultiFilter:
		deserialized_subfilters = []

		for subfilter in value["subfilters"]:
			if subfilter["type"] == FilterType.SINGLE.value:
				subfilter = cast(ExpectedSingleFilterValue, subfilter)
				subfilter_subtype = subfilter["subtype"]

				if subfilter_subtype == FieldType.DATE.value:
					deserializer = DateFilterDeserializer()
				elif subfilter_subtype == FieldType.CHOICE.value:
					deserializer = ChoiceFilterDeserializer()
				elif subfilter_subtype == FieldType.TEXT.value:
					deserializer = TextFilterDeserializer()
				elif subfilter_subtype == FieldType.NUMERIC.value:
					deserializer = NumericFilterDeserializer()
				elif subfilter_subtype == FieldType.BOOLEAN.value:
					deserializer = BooleanFilterDeserializer()
				elif subfilter_subtype == FieldType.ARRAY.value:
					deserializer = ArrayFilterDeserializer()
				else:
					raise NotImplementedError(
						f"Deserialization of subfilters of subtype \"{subfilter['subtype']}\" is not implemented",
					)

				deserialized_subfilters.append(deserializer.deserialize(subfilter))

			elif subfilter["type"] == FilterType.MULTI.value:
				subfilter = cast(ExpectedValue, subfilter)
				deserialized_subfilters.append(self.deserialize(subfilter))

			else:
				raise NotImplementedError(
					f"Deserialization of subfilters of type \"{subfilter['type']}\" is not implemented",
				)

		return MultiFilter(
			clause=Clause(value["clause"]),
			enabled=value["enabled"],
			subfilters=deserialized_subfilters,
		)



from cascade_filter.clause import Clause, CLAUSES_STRICT_COMPARISON, CLAUSES_NON_STRICT_COMPARISON, CLAUSES_NULLABLE
from cascade_filter.exception import ValidationError
from cascade_filter.filter.single.validator import SingleFilterValidator, CleanedData as _CleanedData
from cascade_filter.filter.value.validator import CleanedData as _FilterValueCleanedData

from typing import Final, FrozenSet
from re import Pattern, compile


class FilterValueCleanedData(_FilterValueCleanedData):
	value: str


class CleanedData(_CleanedData):
	value: FilterValueCleanedData


class DateFilterValidator(SingleFilterValidator):
	DATE_RE: Pattern = compile(r"\d\d\d\d-\d\d-\d\d")
	AVAILABLE_CLAUSES: Final[FrozenSet[Clause]] = frozenset((
		*CLAUSES_STRICT_COMPARISON, *CLAUSES_NON_STRICT_COMPARISON, *CLAUSES_NULLABLE,
	))

	def validate(self, value: dict) -> CleanedData:
		base_cleaned_data = super().validate(value)

		if base_cleaned_data["clause"] not in self.AVAILABLE_CLAUSES:
			raise ValidationError("The \"clause\" field must contain one of available clauses")

		filter_value = base_cleaned_data["value"]["value"]

		if not isinstance(filter_value, str) or not self.DATE_RE.match(filter_value):
			raise ValidationError("The \"value\" must be a valid string in format YYYY-MM-DD")

		return base_cleaned_data



from cascade_filter.filter.multi.dataclass import MultiFilter
from cascade_filter.filter.multi.validator import MultiFilterValidator
from cascade_filter.filter.multi.deserializer import MultiFilterDeserializer


__all__ = ["MultiFilter", "MultiFilterValidator", "MultiFilterDeserializer"]



from cascade_filter.clause import Clause
from cascade_filter.filter import ArrayFilter
from cascade_filter.compliance_checker.single import SingleChecker
from cascade_filter.compliance_checker.decorator import nullable


class ArrayChecker(SingleChecker[ArrayFilter]):
	@nullable
	def is_fit(self, obj: object) -> bool:
		attr_value = getattr(obj, self.cascade_filter.table_field, None)

		if attr_value is None:
			return False

		if self.cascade_filter.clause is Clause.CONTAINS:
			return set(attr_value).issuperset(set(self.cascade_filter.value.value))
		elif self.cascade_filter.clause is Clause.NOT_CONTAINS:
			return set(attr_value).isdisjoint(set(self.cascade_filter.value.value))

		raise NotImplementedError(f"Unknown array filter clause: {self.cascade_filter.clause}")



from cascade_filter.filter.numeric.dataclass import NumericFilter
from cascade_filter.filter.numeric.validator import NumericFilterValidator
from cascade_filter.filter.numeric.deserializer import NumericFilterDeserializer


__all__ = ["NumericFilter", "NumericFilterValidator", "NumericFilterDeserializer"]



from cascade_filter.field_type import FieldType
from cascade_filter.filter.single import SingleFilter
from cascade_filter.filter.value import FilterValue

from typing import List, Any


class ArrayFilterValue(FilterValue[List[Any]]):
	pass


class ArrayFilter(SingleFilter):
	subtype = FieldType.ARRAY



from cascade_filter.filter import SingleFilter
from cascade_filter.compliance_checker.base import BaseChecker

from typing import TypeVar, Generic


T = TypeVar("T", bound=SingleFilter)


class SingleChecker(BaseChecker, Generic[T]):
	cascade_filter: T

	def __init__(self, cascade_filter: T) -> None:
		super().__init__(cascade_filter)
